import os
from swepy.pipeline import Swepy
from swepy.nsidcDownloader import nsidcDownloader
import glob
import pytest
import datetime
import numpy as np


def test_login():
    s1 = Swepy(os.getcwd())
    s1.set_login("test", "test")
    assert s1.nD is not None



from affine import Affine
from osgeo import ogr, osr
import re
import sys

resolutions = ["25", "12.5", "6.25", "3.125"]


class EaseReproject:
    gridname = None
    epsg4326Proj4text = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"

    def __init__(self, gridname=None, verbose=False):
        """
        easeReproject.EaseReproject(gridname)

        Returns a transform object for the given gridname.

        Parameters:
            gridname: string
                EASE-Grid 2.0 gridname, following the pattern:
                "EASE2-<proj><res>km", where:
                <proj> is "N", "S", or "T"
                <res> is "25", "12.5", "6.25", or "3.125"
            verbose: bool (optional)
                If True, write verbose output to stderr

        Returns: initialized transformer for gridname

        Example:

        from swepy.easeReproject import EaseReproject
        N25grid = EaseReproject("EASE2_N25km")
        """
        self.gridname = gridname
        g = re.match(r"(EASE2_[NST])([0-9\.]+)km", gridname)
        if g is None:
            print(
                "%s: error parsing gridname %s" % (__name__, gridname),
                file=sys.stderr,
                flush=True,
            )
            raise ValueError
        projection = g.group(1)
        resolution = g.group(2)

        # Check resolution for errors
        if resolution not in resolutions:
            print(
                "%s : unrecognized resolution %s" % (__name__, resolution),
                file=sys.stderr,
                flush=True,
            )
            raise ValueError

        # the geotransform information is the set of GDAL affine transform parameters:
        # (map_UL_x, scale_x, b, map_UL_y, d, scale_y)
        if projection == "EASE2_N":
            self.proj4text = (
                "+proj=laea +lat_0=90 +lon_0=0 "
                + "+x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m"
            )
            self.map_UL_x = -9000000.0
            self.map_UL_y = 9000000.0
            self.b = 0.0
            self.d = 0.0
            self.scale_x = float(resolution) * 1000
            self.scale_y = -1 * float(resolution) * 1000
        elif projection == "EASE2_S":
            self.proj4text = (
                "+proj=laea +lat_0=-90 +lon_0=0 "
                + "+x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m"
            )
            self.map_UL_x = -9000000.0
            self.map_UL_y = 9000000.0
            self.b = 0.0
            self.d = 0.0
            self.scale_x = float(resolution) * 1000
            self.scale_y = -1 * float(resolution) * 1000
        elif projection == "EASE2_T":
            self.proj4text = (
                "+proj=cea +lat_0=0 +lon_0=0 +lat_ts=30 "
                "+x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m"
            )
            self.map_UL_x = -17367530.44
            self.map_UL_y = 6756820.20000
            self.b = 0.0
            self.d = 0.0
            base_resolution_m = 25025.26000
            factor = resolutions.index(resolution)
            self.scale_x = base_resolution_m / (2.0 ** factor)
            self.scale_y = -1 * base_resolution_m / (2.0 ** factor)
        else:
            print(
                "%s : unrecognized projection %s" % (__name__, projection),
                file=sys.stderr,
                flush=True,
            )
            raise ValueError

        geotransform = (
            self.map_UL_x + self.scale_x / 2.0,
            self.scale_x,
            self.b,
            self.map_UL_y + self.scale_y / 2.0,
            self.d,
            self.scale_y,
        )
        self.fwd = Affine.from_gdal(*geotransform)

        # Initialize / Save coordinate transform for this projection
        self.gridSpatialRef = osr.SpatialReference()
        self.gridSpatialRef.SetFromUserInput(self.proj4text)

        # Initialize and save coordinate transformation for EPSG4326 (lat/lon)
        self.epsg4326SpatialRef = osr.SpatialReference()
        self.epsg4326SpatialRef.SetFromUserInput(self.epsg4326Proj4text)

        # Initialize and save the forward and reverse transformations
        self.projToGeog = osr.CoordinateTransformation(
            self.gridSpatialRef, self.epsg4326SpatialRef
        )
        self.geogToProj = osr.CoordinateTransformation(
            self.epsg4326SpatialRef, self.gridSpatialRef
        )

        if verbose:
            print(
                "%s : initialized new Ease2Transform object" % (__name__),
                file=sys.stderr,
                flush=True,
            )

    def grid_to_map(self, row, col):
        """
        swepy.easeReproject.EaseReproject.grid_to_map(row, col)

        Parameters: row, col : scalars
                        grid locations to convert, grid origin defined as
                        (row, col) = (0., 0.) at center of UL grid cell

        Returns: (x, y) map coordinates in meters

        Example:

        from swepy.easeReproject import EaseReproject

        N25grid = EaseReproject("EASE2_N25km")
        (x, y) = N25grid.grid_to_map(-0.5, -0.5)

        Returns (x, y) = (-9000000., 9000000), UL corner of UL cell
        """
        ax, ay = self.fwd * (col, row)
        return (ax, ay)

    def grid_to_geographic(self, row, col):
        """
        swepy.easeReproject.EaseReproject(row, col)

        Parameters: row, col : scalars
                        grid locations to convert, grid origin defined as
                        (row, col) = (0., 0.) at center of UL grid cell

        Returns: (lat, lon) geographic coordinates in degrees

        Example:

        from swepy.easeReproject import EaseReproject

        N25grid = EaseReproject("EASE2_N25km")
        (lat, lon) = N25grid.grid_to_geographic(359.5, 359.5)

        Returns (lat, lon) = (90., 0.), North pole
        """
        # get map coordinates of row, col
        x, y = self.grid_to_map(row, col)

        # Create a geometry with the map coordinates
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(x, y)

        point.Transform(self.projToGeog)

        return (point.GetY(), point.GetX())

    def map_to_grid(self, x, y):
        """
        swepy.easeReproject.EaseReproject.map_to_grid(x, y)

        Parameters: x, y : scalars
                        map locations (in meters) to convert
        Returns: (row, col) grid coordinates, grid origin defined as
                        (row, col) = (0., 0.) at center of UL grid cell

        Examples:

        from swepy.easeReproject import EaseReproject

        N25grid = EaseReproject("EASE2_N25km")
        (row, col) = N25grid.map_to_grid(-9000000., 9000000.)

        Returns (row, col) = (-0.5, -0.5), UL corner of UL cell
        """
        col, row = ~self.fwd * (x, y)
        return (row, col)

    def geographic_to_grid(self, lat, lon):
        """
        swepy.easeReproject.EaseReproject.geographic_to_grid(lat, lon)

        Parameters: lat, lon : scalars
                        geographic coordinates (in degrees) to convert

        Returns: (row, col) converted grid location, grid origin defined as
                        (row, col) = (0., 0.) at center of UL grid cell

        Example:

        from swepy.easeReproject import EaseReproject
        N25grid = EaseReproject("EASE2_N25km")
        (row, col) = N25grid.geographic_to_grid(90., 0.)

        Returns (row, col) = (359.5, 359.5), grid location of North Pole
        """
        # create a geom with the geographic coordinates
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(lon, lat)
        point.Transform(self.geogToProj)

        # returned values are in meters, convert to grid
        row, col = self.map_to_grid(point.GetX(), point.GetY())
        return (row, col)



import datetime
import time
import os
import warnings
import numpy as np
from numpy import newaxis
from numpy import shape
import pandas as pd
from skimage.measure import block_reduce
import math
from scipy.signal import savgol_filter
from scipy.cluster.vq import *
import jenkspy
import numpy.ma as ma
from netCDF4 import Dataset
from multiprocessing import Pool, Process, cpu_count


def get_array(file):
    """
    Take 19H and 37H netCDF files, open and store tb
    data in np arrays

    Parameters
    -----------
    file19: str
        filename for 19H file
    file37: str
        filename for 37H file
    high: bool
        true = high resolution imagery (3.125km/6.25km)
        false = low resolution imagery (25km)
    """
    fid = Dataset(file, "r", format="NETCDF4")
    tb = fid.variables["TB"][:]
    if fid.variables["crs"].long_name == "EASE2_N3.125km":
        tb[tb.mask] = 0.00001
        tb = block_reduce(tb, block_size=(1, 2, 2), func=np.mean)
        fid.close()
        return ma.masked_values(tb, 0.00001)
    else:
        fid.close()
        return tb


def pandas_fill(arr):
    """
    Given 2d array, convert to pd dataframe
    and ffill missing values in place

    Parameters
    ----------
    arr: np.array
        Ideally time vector of swe cube
    """
    df = pd.DataFrame(arr)
    df.fillna(method="ffill", inplace=True)
    out = df.values
    return out


def vector_clean(cube):  # rename to ffill_array(cube)
    """
    Clean erroneous spikes out of 37Ghz cube

    Parameters
    -----------
    cube: np.array(t,x,y)
        np array time cube of 37GHz tb data
    Note: "cube" can be used with other arrays but is looking for patterns in 37H files
    """
    cube[cube == 0] = np.nan
    for i in range(np.shape(cube)[0]):
        arr = cube[i, :, :]
        mask = np.isnan(arr)
        idx = np.where(~mask, np.arange(mask.shape[1]), 0)
        np.maximum.accumulate(idx, axis=1, out=idx)
        cube[i, :, :] = arr[np.arange(idx.shape[0])[:, None], idx]
    return cube


def apply_filter(cube):
    """
    Apply a sav-gol filter from scipy to time vector's of cube

    Parameters
    -----------
    cube: np.array(t,x,y)
        np array time cube of swe for passive microwave data
    """
    shapecube = np.shape(cube)
    smooth_cube = np.empty((shapecube[0], shapecube[1], shapecube[2]))
    if shapecube[0] == 1:
        print("Cannot smooth a cube with time vector of length 1.")
        return ValueError
    elif (
        shapecube[0] < 51
    ):  # when time vector is len(1) --> concat over itself to make len(3)
        window = shapecube[0] - 1 if shapecube[0] % 2 == 0 else shapecube[0]
        poly = 3 if window > 3 else window - 1
    else:
        window = 51
        poly = 3
    for x in range(shapecube[1]):
        for y in range(shapecube[2]):
            pixel_drill = cube[:, x, y]
            pixel = pandas_fill(pixel_drill)
            yhat = savgol_filter(np.squeeze(pixel), window, poly)
            yhat[yhat < 2] = 0
            smooth_cube[:, x, y] = yhat
    return smooth_cube


def apply_filter_mphelper(cube):
    """
    Helper function to apply the filter function in a parralel fashion
    Makes use of a Pool to process on every available core

    Parameters
    -----------
    cube: np.array
        numpy array of data, should be 3d (x,x,x)
    """
    cpus = cpu_count()
    try:
        swe_parts = np.array_split(cube, cpus, axis=2)
    except IndexError:
        print(
            "Array Provided does not have a 2nd axis to split on. Please provide a 3 dimensional cube."
        )
    with Pool(cpus) as p:
        parts = p.map(apply_filter, swe_parts)
        try:
            return np.concatenate(parts, axis=2)  # recombine split cube
        except ValueError:
            # FIND WAY TO GET HERE VIA TEST SUITE (1 CORE)
            print(
                "Array provided is smaller than # of cores available. Exiting"
            )


def auto_filter(file19, file37):  # filter_swe is either filter on tb or swe
    """
    Clean missing values and apply sav gol filter, return SWE cube

    Parameters
    -----------
    cube19: np.array
        size = (x,x,x)
    cube37: np.array
        size = (x,x,x)
    """
    cube19, cube37 = get_array(file19, file37)
    clean19 = vector_clean(cube19)
    clean37 = vector_clean(cube37)
    swe = safe_subtract(clean19, clean37)
    return apply_filter_mphelper(swe)


def govf(array, classes):
    # get break points
    classes = jenkspy.jenks_breaks(array, classes)
    # do classificaton
    classified = np.array([classify(i, classes) for i in array])
    # max value of zones
    maxz = max(classified)
    # nested list of zone indices
    zone_indices = [
        [idx for idx, val in enumerate(classified) if zone + 1 == val]
        for zone in range(maxz)
    ]
    # sum of squared deviations from array mean
    sdam = np.sum((array - array.mean()) ** 2)
    # sorted polygon stats
    array_sort = [
        np.array([array[index] for index in zone]) for zone in zone_indices
    ]
    # sum of squared deviations of class means
    sdcm = sum(
        [
            np.sum((classified - classified.mean()) ** 2)
            for classified in array_sort
        ]
    )
    # goodness of variance fit
    gvf = (sdam - sdcm) / sdam
    return gvf


def classify(value, breaks):
    for i in range(1, len(breaks)):
        if value < breaks[i]:
            return i
    return len(breaks) - 1


def mask_ocean_winter(swe_matrix, day=0, nclasses=3):
    """
    Use a winter day to mask ocean pixels out of coastal imagery in arctic.

    There is a clear difference between winter land pixels and ocean pixels
    that classification can sort out for us using a simple jenks classification.
    Data should have already moved through "vector_clean" and "apply_filter"

    Parameters
    ----------
    swe_matrix: np.array
        swe time cube
    day: int
        julian date of time series to use for classification (should be winter)
    nclasses: int
        number of classes to use in jenks classification, defaults to 3
    """
    winter_day = swe_matrix[day, :, :]
    classes_jenk = jenkspy.jenks_breaks(winter_day.ravel(), nclasses)
    mask = classes_jenk == 1
    winter_day[mask] = -8888
    matrix_mask = np.zeros(swe_matrix.shape, dtype=bool)
    matrix_mask[:, :, :] = winter_day[np.newaxis, :, :] == -8888
    swe_matrix[matrix_mask] = -8888
    return swe_matrix


def safe_subtract(tb19, tb37):
    """
    Check size of each file, often the 19 and 37
    matrices are one unit off of eachother.

    Chops the larger matrix to match the smaller matrix
    """

    shape1 = np.shape(tb19)
    shape2 = np.shape(tb37)
    s1 = [shape1[0], shape1[1], shape1[2]]
    s2 = [shape2[0], shape2[1], shape2[2]]
    if s1[1] < s2[1]:
        s2[1] = s1[1]
    elif s1[1] > s2[1]:
        s1[1] = s2[1]
    if s1[2] < s2[2]:
        s2[2] = s1[2]
    elif s1[2] > s2[2]:
        s1[2] = s2[2]
    tb19 = tb19[:, : s1[1] - 1, : s1[2] - 1]
    tb37 = tb37[:, : s2[1] - 1, : s2[2] - 1]
    tb = tb19 - tb37
    return tb



# Unit Testing for SWEpy
import os
from swepy.pipeline import Swepy
from swepy.nsidcDownloader import nsidcDownloader
import glob
import pytest
import datetime
import numpy as np
from shutil import copy


def set_login_test(self, username="test", password="test"):
    """
        Monkey patched set_login function
    """
    self.local_session = True
    self.nD = nsidcDownloader(folder=os.getcwd(), no_auth=True)
    self.nD.username = "test"
    self.nD.password = "test"
    self.username = username
    self.password = password


Swepy.set_login = set_login_test


def test_nodir():
    s1 = Swepy(None)
    assert s1.working_dir == os.getcwd()


def test_check_params_false():
    ul = [66, -145]
    s1 = Swepy(os.getcwd(), ul, ul)
    assert s1.check_params() is False


def test_check_params_true():
    start = datetime.date(2010, 1, 1)
    ul = [66, -145]
    lr = [76, -166]
    s1 = Swepy(os.getcwd(), ul, lr)
    s1.set_login("test", "test")
    s1.set_dates(start, start)
    assert s1.check_params() is True


def test_set_params_bounds():
    start = datetime.date(2010, 1, 1)
    s1 = Swepy(os.getcwd())
    s1.set_dates(start, start)
    s1.set_login("test", "test")
    s1.set_grid(ul=[66, -145], lr=[73, -166])
    assert s1.check_params() is True


def test_set_params_auth():
    start = datetime.date(2010, 1, 1)
    s1 = Swepy(os.getcwd(), ul=[66, -145], lr=[73, -166])
    s1.set_dates(start, start)
    s1.set_login(username="test", password="test")
    assert s1.check_params() is True


def test_set_params_dates():
    s1 = Swepy(os.getcwd(), ul=[66, -145], lr=[73, -166])
    s1.set_login("test", "test")
    s1.set_dates(
        start=datetime.date(2010, 1, 1), end=datetime.date(2010, 1, 1)
    )
    assert s1.check_params() is True


def test_get_grid_N():
    s1 = Swepy(os.getcwd())
    lat1 = 45
    lat2 = 80
    assert s1.get_grid(lat1, lat2) == "N"


def test_get_grid_T():
    s1 = Swepy(os.getcwd())
    lat1 = 30
    lat2 = 30
    assert s1.get_grid(lat1, lat2) == "T"


def test_get_grid_S():
    s1 = Swepy(os.getcwd())
    lat1 = -50
    lat2 = -80
    assert s1.get_grid(lat1, lat2) == "S"


def test_get_grid_fails():
    s1 = Swepy(os.getcwd())
    lat1 = -62
    lat2 = 70
    with pytest.raises(Exception):
        s1.get_grid(lat1, lat2)


def test_get_xy_N():
    ll_ul = [62, -140]
    ll_lr = [73, -166]
    s1 = Swepy(os.getcwd(), ul=ll_ul, lr=ll_lr)
    list1 = s1.get_xy(ll_ul, ll_lr)
    assert list1 == pytest.approx(
        [
            -1988822.7284991555,
            2370186.631721887,
            -457544.84080317616,
            1835112.1237310776,
        ],
        0.1,
    )


def test_get_xy_S():
    ll_lr = [-80, 9]
    ll_ul = [-69, -16]
    s1 = Swepy(os.getcwd(), ul=ll_ul, lr=ll_lr)
    list1 = s1.get_xy(ll_ul, ll_lr)
    assert list1 == pytest.approx(
        [
            -642633.6942027323,
            2241130.027261451,
            174488.41818780638,
            1101676.5146265049,
        ],
        0.1,
    )


def test_get_xy_none():
    ll_lr = None
    ll_ul = None
    s1 = Swepy(os.getcwd())
    with pytest.raises(Exception):
        s1.get_xy(ll_ul, ll_lr)


def test_get_file_high():
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login()
    date = datetime.datetime(2010, 1, 1)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "6.25km",
        "platform": "F17",
        "sensor": "SSMIS",
        "date1": datetime.datetime(2010, 1, 1),
        "date2": datetime.datetime(2010, 1, 1),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "SIR",
    }


def test_get_file_low():
    s1 = Swepy(os.getcwd(), ul="N", lr="N", high_res=False)
    s1.set_login()
    date = datetime.datetime(2010, 1, 1)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "25km",
        "platform": "F17",
        "sensor": "SSMIS",
        "date1": datetime.datetime(2010, 1, 1),
        "date2": datetime.datetime(2010, 1, 1),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "GRD",
    }


def test_gf_low_var1():
    s1 = Swepy(os.getcwd(), ul="N", lr="N", high_res=False)
    s1.set_login()
    date = datetime.datetime(2003, 11, 6)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "25km",
        "platform": "F14",
        "sensor": "SSMI",
        "date1": datetime.datetime(2003, 11, 6),
        "date2": datetime.datetime(2003, 11, 5),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "GRD",
    }


def test_gf_low_var2():
    s1 = Swepy(os.getcwd(), ul="N", lr="N", high_res=False)
    s1.set_login()
    date = datetime.datetime(2006, 11, 4)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "25km",
        "platform": "F15",
        "sensor": "SSMI",
        "date1": datetime.datetime(2006, 11, 4),
        "date2": datetime.datetime(2006, 11, 3),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "GRD",
    }


def test_gf_low_TA():
    s1 = Swepy(os.getcwd(), ul="T", lr="T", high_res=False)
    s1.set_login()
    date = datetime.datetime(2006, 11, 4)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "25km",
        "platform": "F15",
        "sensor": "SSMI",
        "date1": datetime.datetime(2006, 11, 4),
        "date2": datetime.datetime(2006, 11, 3),
        "channel": "19H",
        "grid": "T",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "A",
        "algorithm": "GRD",
    }


def test_df_smmr():
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login()
    date = datetime.datetime(1979, 1, 1)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "6.25km",
        "platform": "NIMBUS7",
        "sensor": "SMMR",
        "date1": datetime.datetime(1979, 1, 1),
        "date2": datetime.datetime(1979, 1, 1),
        "channel": "18H",
        "grid": "N",
        "input": "JPL",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "SIR",
    }


def test_gs_1987():
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login()
    date = datetime.datetime(1987, 8, 21)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "6.25km",
        "platform": "F08",
        "sensor": "SSMI",
        "date1": datetime.datetime(1987, 8, 21),
        "date2": datetime.datetime(1987, 8, 21),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "SIR",
    }


def test_gs_2008_edge():
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login()
    date = datetime.datetime(2008, 3, 7)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "6.25km",
        "platform": "F16",
        "sensor": "SSMIS",
        "date1": datetime.datetime(2008, 3, 7),
        "date2": datetime.datetime(2008, 3, 7),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "M",
        "algorithm": "SIR",
    }


def test_gs_evening_edge():
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login()
    date = datetime.datetime(2005, 5, 12)
    file = s1.get_file(date, "19H")
    assert file == {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "resolution": "6.25km",
        "platform": "F15",
        "sensor": "SSMI",
        "date1": datetime.datetime(2005, 5, 12),
        "date2": datetime.datetime(2005, 5, 12),
        "channel": "19H",
        "grid": "N",
        "input": "CSU",
        "dataversion": "v1.3",
        "pass": "E",
        "algorithm": "SIR",
    }


def test_safe_subtract1():
    s1 = Swepy(os.getcwd())
    tb19 = np.ones((1, 152, 153))
    tb37 = np.ones((1, 154, 153))
    tb = s1.safe_subtract(tb19, tb37)
    assert np.shape(tb) == (1, 151, 152)


def test_safe_subtract2():
    s1 = Swepy(os.getcwd())
    tb19 = np.ones((1, 154, 152))
    tb37 = np.ones((1, 152, 153))
    tb = s1.safe_subtract(tb19, tb37)
    assert np.shape(tb) == (1, 151, 151)


def test_scrape_fail():
    list1 = glob.glob("*.nc")
    assert list1 != [
        "NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010002-37H-M-SIR-CSU-v1.3.nc",
        "NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010002-19H-M-SIR-CSU-v1.3.nc",
    ]


def test_scrape():
    date = datetime.date(2010, 1, 1)
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_login("test", "test")
    s1.set_dates(date, date)
    s1.scrape()
    list1 = glob.glob("*.nc")
    assert list1 == [
        "NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc",
        "NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc",
    ] or list1 == [
        "NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc",
        "NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc",
    ]


def test_subset():
    if not os.path.exists("sub"):
        os.mkdir("sub")
    s1 = Swepy(os.getcwd(), ul=[66, -145], lr=[71, -166])
    setattr(
        s1,
        "down19list",
        ["NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc"],
    )
    setattr(
        s1,
        "down37list",
        ["NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc"],
    )
    path = os.getcwd() + "/"
    # '/tests/data/MEASURES/NSIDC-0630.001/2010.01.01/'
    out_dir = path + "/sub/"
    s1.subset(in_dir=path, out_dir19=out_dir, out_dir37=out_dir)
    os.chdir(out_dir)
    list1 = glob.glob("*.nc")
    assert os.stat(list1[0]).st_size < 100000


def test_concat():
    s1 = Swepy(os.getcwd(), ul=[66, -145], lr=[73, -166])
    setattr(
        s1,
        "sub19list",
        ["NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc"],
    )
    setattr(
        s1,
        "sub37list",
        ["NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc"],
    )
    s1.concatenate()
    list1 = glob.glob("*all*")
    assert "all_days_19H.nc" in list1


# def test_concat_subFalse():
#     s1 = swepy(os.getcwd(), ul='N',lr='N')
#     copy('NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc', '/wget/NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc')
#     copy('NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc', '/wget/NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc')
#     setattr(s1,'down19list',['NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc'])
#     setattr(s1,'down37list',['NSIDC-0630-EASE2_N3.125km-F17_SSMIS-2010001-37H-M-SIR-CSU-v1.3.nc'])
#     s1.concatenate()


def test_clean_dirs():
    s1 = Swepy(os.getcwd())
    try:
        s1.clean_dirs()
        list1 = glob.glob("*nc")
        assert list1 == []
    except PermissionError:
        assert True


def test_final_concat_fail():
    s1 = Swepy(os.getcwd())
    res = s1.final_concat()
    assert res == ("all_days_19H.nc", "all_days_37H.nc")


# def test_scrape_all():
#     date = datetime.datetime(2010,1,1)
#     s1 = swepy(os.getcwd(), start = date, end = date,ul = 'N', lr = 'N', username = 'test', password='test')
#     res = s1.scrape_all()
#     assert res == ['all_days_19H.nc', 'all_days_37H.nc']



# author: Will Norris --> wino6687@colorado.edu, Earth Lab, CU Boulder
from datetime import datetime, timedelta
import swepy.nsidcDownloader as nsidcDownloader
import swepy.easeReproject as easeReproject
import numpy as np
from netCDF4 import Dataset
import pandas as pd
from nco import Nco
import os
from tqdm import tqdm
import glob
import xarray
import fsspec
from mapboxgl.utils import create_color_stops, df_to_geojson
from mapboxgl.viz import CircleViz
import numpy.ma as ma


nco = Nco()


class Swepy:
    """
    Class to facilitate the scraping/subsetting/concatenating of tB files for SWE analysis.
    """

    def __init__(
        self,
        working_dir=None,
        ul=None,
        lr=None,
        outfile19="all_days_19H.nc",
        outfile37="all_days_37H.nc",
        high_res=True,
    ):
        """
        Parameters
        ----------
        working_dir: str
            directory to store data directory
        start: datetime
            start date for scraping
        end: datetime
            end date for scraping
        ul: List[int, int]
            upper left bounding coordinates [lat, lon]
        lr: List[int, int]
            lower right bounding coordinates [lat,lon]
        username: str
            username for Earth Data login
        password: str
            password for Earth Data login
        outfile19: str
            name of final output file, 19 19GHz
        outfile37: str
            name of final output file, 37 GHz
        high_res: bool
            True: scrape high resolution files, False: low resolution
        """

        # set whether we are scraping resampled data or not
        if high_res:
            self.high_res = True
        else:
            self.high_res = False

        # set the working directory
        if working_dir is None:
            print(
                "No working directory passed, using current working directory: {}".format(
                    os.getcwd()
                )
            )
            self.working_dir = os.getcwd()
        else:
            self.working_dir = working_dir

        # create directories for data
        self.wget, self.path19, self.path37 = self.get_directories(
            self.working_dir
        )

        self.outfile_19 = outfile19
        self.outfile_37 = outfile37
        self.ease = None
        self.geo_list = None
        self.grid = None
        self.username = None
        self.password = None
        self.dates = None

        self.set_grid(ul, lr)

        self.down19list = []
        self.down37list = []
        self.sub19list = []
        self.sub37list = []
        self.concatlist = [None, None]
        self.concat19list = []
        self.concat37list = []

    def set_dates(self, start=None, end=None):
        """
        Set date range using start and end datetime objects

        Parameters
        ----------
        start: datetime
            start date for scraping
        end: datetime
            end date for scraping
        """

        try:
            self.dates = pd.date_range(start, end)
            print("Dates set successfully...")
        except ValueError:
            self.dates = None
            print("No valid dates given")
        return

    def set_login(self, username=None, password=None):
        """
        Set login credentials and login to earth data

        Parameters
        ----------
        username: String
            Earthdata username
        password: String
            Earthdata password
        """

        if username is not None and password is not None:
            print("Logging you into Earth Data...")
            self.local_session = False
            self.nD = nsidcDownloader.nsidcDownloader(
                folder=self.wget, username=username, password=password
            )
            print("Successfully logged in!")
            self.nD.username = username
            self.nD.password = password
            self.username = username
            self.password = password
        else:
            print(
                "No credentials given, please use 'set_login' to login when ready."
            )
            self.nD = None
        return

    def set_grid(self, ul=None, lr=None):
        """
        Set grid corners, and convert to xy

        Parameters
        ----------
        ul: char or [float,float]
            upper left bounding coordinates or grid name (N,S,T)
        lr: [float, float]
            lower right bounding coordinates (not needed for entire grid)
        """

        if ul is not None and lr is not None:
            # if ul is an entire grid, that is the grid we want to scrape
            if ul in ["N", "S", "T"]:
                self.grid = ul
                self.subBool = False
                self.geo_list = ul
            else:  # if specific coords, we want to get the x_y ease coords that correspond
                self.subBool = True
                self.get_grid(ul[0], lr[0])
                self.geo_list = self.get_xy(ul, lr)
                self.center = [ul[1], ul[0]]
        else:
            print(
                "No usable bounding coordinates or grid given, \
             please specify bounds or a grid to scrape."
            )
        return

    def get_grid(self, lat1, lat2):
        """
        Check which regions the lats fall into. Based on the grid, instantiate the ease
        grid conversion object.

        Parameters
        ----------
        lat1: int
            Upper Left Latitude
        lat2: int
            Lower Right Latitude
        """

        if (lat1 < 50 and lat2 < 50) and (
            lat1 > -50 and lat2 > -50
        ):  # mid lat
            self.grid = "T"
            self.ease = easeReproject.EaseReproject("EASE2_T3.125km")
        elif (lat1 > 40 and lat2 > 40) and (lat1 < 90 and lat2 < 90):  # north
            self.grid = "N"
            self.ease = easeReproject.EaseReproject("EASE2_N3.125km")
        elif (lat1 < -40 and lat2 < -40) and (
            lat1 > -90 and lat2 > -90
        ):  # South
            self.grid = "S"
            self.ease = easeReproject.EaseReproject("EASE2_S3.125km")
        else:
            print(
                "SWEpy currently only supports subsetting study areas with a study area in the North, South, or Equatorial imagery \
            \nOverlappig study areas can cause errors, and require more than one region of imagery"
            )
            raise ValueError
        return self.grid

    def get_directories(self, path):
        """
        Given a working directory, create data
        directories if non-existent

        Parameters
        ----------
        path: str
            working directory to create data directories
        """
        os.chdir(path)
        paths = [
            path + "/data/wget/",
            path + "/data/Subsetted_19H/",
            path + "/data/Subsetted_37H/",
        ]
        for path1 in paths:
            if not os.path.exists(path1):
                os.makedirs(path1)
        return paths[0], paths[1], paths[2]

    def get_xy(self, ll_ul, ll_lr):
        """
        Use nsidc scripts to convert user inputted
        lat/lon into Ease grid 2.0 coordinates

        Parameters
        -----------

        ll_ul: [float, float]
            Latitude and longitude for upper left coordinates
        ll_lr: [float, float]
            Latitude and longitude for lower right coordinates
        """

        if ll_ul is None or ll_lr is None:
            print(
                "You must enter bounding coordinates, please use 'set_grid()' to add coordinates"
            )
            raise ValueError
        row, col = self.ease.geographic_to_grid(ll_ul[0], ll_ul[1])
        xul, yul = self.ease.grid_to_map(row, col)
        row, col = self.ease.geographic_to_grid(ll_lr[0], ll_lr[1])
        xlr, ylr = self.ease.grid_to_map(row, col)
        return [xul, yul, xlr, ylr]

    def subset(
        self, scrape=False, in_dir=None, out_dir19=None, out_dir37=None
    ):
        """
        Get the files from wget directory
        and subset them geographically based on
        coords from constructor

        Parameters
        ----------

        scrape: Boolean
            Under the hood variable to allow for auto workflow

        in_dir: str
            (Optional) directory with wget data stored in it.
            Default: "working_dir/data/wget"

        out_dir19: str
            (Optional) directory to store output 19GHz files
            Default: "working_dir/data/Subsetted_19H/"

        out_dir37: str
            (Optional) directory to store output 37GHz files
            Default: "working_dir/data/Subsetted_37H"
        """

        os.chdir(self.working_dir + "/data")
        for file in tqdm(self.down19list):
            outfile = (
                self.path19 + file if out_dir19 is None else out_dir19 + file
            )
            infile = self.wget + file if in_dir is None else in_dir + file
            opt = [
                "-d x,%f,%f" % (self.geo_list[0], self.geo_list[2]),
                "-d y,%f,%f" % (self.geo_list[3], self.geo_list[1]),
                "-v TB",
            ]
            nco.ncks(input=infile, output=outfile, options=opt)
            self.sub19list.append(outfile)
            os.remove(infile)

        for file in tqdm(self.down37list):
            outfile = (
                self.path37 + file if out_dir37 is None else out_dir37 + file
            )
            infile = self.wget + file if in_dir is None else in_dir + file
            opt = [
                "-d x,%f,%f" % (self.geo_list[0], self.geo_list[2]),
                "-d y,%f,%f" % (self.geo_list[3], self.geo_list[1]),
                "-v TB",
            ]
            nco.ncks(input=infile, output=outfile, options=opt)
            self.sub37list.append(outfile)
            os.remove(infile)
        self.down19list = []
        self.down37list = []
        return

    '''
    def scrape_all(self):
        """
        Function to parralelize scraping with nsidcDownloader

        generate complete list of files to scrape, then divide into process pool
    '''

    def scrape_all(self):  # rename to be related to 'full workflow'
        """
        Function to ensure we subset and concatenate every year!
        Implements the whole workflow!
        """
        # can't scrape all unless all parameters entered
        if self.check_params() is False:
            return None, None
        if len(self.dates) <= 300:
            self.scrape()
            if self.subBool:
                self.subset()
            return self.concatenate()
        else:
            comp_list = [
                self.dates[x : x + 300] for x in range(0, len(self.dates), 300)
            ]
            for count, subList in enumerate(comp_list):
                name19 = "temp19_" + str(count) + ".nc"
                name37 = "temp37_" + str(count) + ".nc"
                self.scrape(subList)
                if self.subBool:
                    self.subset()
                self.concatenate(
                    name19, name37, all=True
                )  # CAN I PARALLELIZE HERE
            return self.final_concat()

    def convert_netcdf_zarr(self, outname19="zarr19", outname37="zarr37"):
        """
        Convert netCDF files into zarr directories for storage in S3

        Parameters
        ----------
        outname19: string (optional)
            name of the directory to store 19H file
        outname37: string (optional)
            name of the directory to store 37H file

        Returns
        -------
        dict: {outname19:zarr_obj, outname37:zarr_obj}
            dictionary with filename as key and the zarr object generated as the value
        """
        ds19 = xarray.open_dataset(self.outfile_19, engine="netcdf4")
        ds37 = xarray.open_dataset(self.outfile_37, engine="netcdf4")
        self.zarr19 = ds19.to_zarr(outname19, "w")
        self.zarr37 = ds37.to_zarr(outname37, "w")
        return {outname19: self.zarr19, outname37: self.zarr37}

    def get_sensor(self, date):
        """
        Helper function to return optimal sensor for a given date

        Parameters
        -----------
        date: datetime.date()
            date to find sensor information
        """
        sensors = {
            1978: "NIMBUS7",
            1979: "NIMBUS7",
            1980: "NIMBUS7",
            1981: "NIMBUS7",
            1982: "NIMBUS7",
            1983: "NIMBUS7",
            1984: "NIMBUS7",
            1985: "NIMBUS7",
            1986: "NIMBUS7",
            1987: "NIMBUS7",
            1988: "F08",
            1989: "F08",
            1990: "F08",
            1991: "F10",
            1992: "F11",
            1993: "F11",
            1994: "F11",
            1995: "F11",
            1996: "F13",
            1997: "F13",
            1998: "F13",
            1999: "F13",
            2000: "F13",
            2001: "F13",
            2002: "F13",
            2003: "F15",
            2004: "F15",
            2005: "F15",
            2006: "F16",
            2007: "F16",
            2008: "F16",
            2009: "F17",
            2010: "F17",
            2011: "F17",
            2012: "F17",
            2013: "F17",
            2014: "F18",
            2015: "F19",
            2016: "F18",
        }

        sensor_dict = {
            "F16": "SSMIS",
            "F17": "SSMIS",
            "F18": "SSMIS",
            "F19": "SSMIS",
            "NIMBUS7": "SMMR",
            "F08": "SSMI",
            "F10": "SSMI",
            "F11": "SSMI",
            "F13": "SSMI",
            "F14": "SSMI",
            "F15": "SSMI",
        }

        sensor = sensors[date.year]
        ssmi_s = sensor_dict[sensor]
        # NIMBUS drops off on aug 21, 1987 and F08 starts
        if datetime(1987, 8, 21) <= date <= datetime(1988, 1, 1):
            sensor = "F08"
            ssmi_s = "SSMI"

        if datetime(1991, 3, 10) == date:
            sensor = "F08"

        if not self.high_res:
            if date in [datetime(2003, 11, 6), datetime(2004, 4, 9)]:
                sensor = "F14"
            if date in [
                datetime(2006, 11, 4),
                datetime(2006, 12, 1),
                datetime(2008, 2, 26),
            ]:
                sensor = "F15"
                ssmi_s = "SSMI"
            if date in pd.date_range(
                datetime(2008, 3, 6), datetime(2008, 12, 31)
            ):
                sensor = "F17"

        return sensor, ssmi_s

    def get_file(self, date, channel):  # add more defaulting params
        """
        Function that uses date and channel to find optimal file composition
        and return the file params for the web scraper's use.
        Parameters
        ----------
        date: datetime
            date to find file path for
        channel: str
            19H vs 37H channel
        """
        # JPL curated the data up to (1987,8,21) then CSU
        input_org = "JPL" if date < datetime(1987, 8, 21) else "CSU"
        # there is no 19H on NIMBUS7, scrape 18H (similar)
        if channel == "19H" and date < datetime(1987, 8, 21):
            channel = "18H"

        # determine if high res or not
        if self.high_res:
            resolution = "6.25km" if channel in ["19H", "18H"] else "3.125km"
            algorithm = "SIR"
            date2 = date
        else:  # low res data has problems with file paths, these were manually fixed
            resolution = "25km"
            algorithm = "GRD"
            if datetime(2003, 1, 1) <= date <= datetime(2008, 3, 5):
                date2 = date - timedelta(days=1)
            else:
                date2 = date

        if self.grid == "T":
            pass1 = "A"
        elif date in [
            datetime(2005, 5, 12),
            datetime(2006, 2, 4),
            datetime(2008, 1, 2),
            datetime(2008, 2, 26),
        ]:
            pass1 = "E"
        else:
            pass1 = "M"

        sensor, ssmi_s = self.get_sensor(date)
        file = {
            "protocol": "http" if self.local_session else "https",
            "server": "localhost:8000"
            if self.local_session
            else "n5eil01u.ecs.nsidc.org",
            "datapool": "MEASURES",
            "resolution": resolution,
            "platform": sensor,
            "sensor": ssmi_s,
            "date1": date,
            "date2": date2,
            "channel": channel,
            "grid": self.grid,
            "dataversion": "v1.3",
            "pass": pass1,
            "algorithm": algorithm,
            "input": input_org,
        }
        return file

    def concatenate(self, outname19=None, outname37=None, all=False):
        """
        Function to concatenate files in the subsetted data folders.
        Input parameter is simply to allow for nesting of functions.

        Parameters
        ----------
        outname19 : str
            output file name for 19Ghz
        outname37 : str
            output file name for 37GHz
        all : Boolean
        """

        if self.subBool is False:
            self.sub19list = (
                self.down19list
            )  # sublist is getting reset back to the full download list
            self.sub37list = self.down37list
            os.chdir(self.wget)
        outname19 = self.outfile_19 if outname19 is None else outname19
        outname37 = self.outfile_37 if outname37 is None else outname37
        # Concatenate 19GHz files:
        if len(self.sub19list) != 0:
            nco.ncrcat(input=self.sub19list, output=outname19, options=["-O"])
            for file in self.sub19list:
                os.remove(file)
            self.sub19list = []
            self.down19list = []
            if all:
                self.concat19list.append(outname19)
            else:
                self.concatlist[0] = outname19
        else:
            print("No 19Ghz Files to Concatenate")
        # Concatenate 37GHz files:
        if len(self.sub37list) != 0:
            nco.ncrcat(input=self.sub37list, output=outname37, options=["-O"])
            for file in self.sub37list:
                os.remove(file)
            self.sub37list = []
            self.down37list = []
            if all:
                self.concat37list.append(outname37)
            else:
                self.concatlist[1] = outname37
        else:
            print("No 37Ghz Files to Concatenate")
        return outname19, outname37

    def final_concat(self):
        """
        Manage the final concatenation for scrape_all
        """
        if len(self.concat19list) != 0:
            nco.ncrcat(
                input=self.concat19list, output=self.outfile_19, options=["-O"]
            )
            self.concatlist[0] = self.outfile_19
        else:
            print("No 19Ghz Files to Concatenate")
        if len(self.concat37list) != 0:
            nco.ncrcat(
                input=self.concat37list, output=self.outfile_37, options=["-O"]
            )
            self.concatlist[1] = self.outfile_37
        else:
            print("No 37Ghz Files to Concatenate")

        # clean out files that were concatenated
        for file in self.concat19list:
            os.remove(file)
        for file in self.concat37list:
            os.remove(file)
        self.concat19list = []
        self.concat37list = []
        return self.outfile_19, self.outfile_37

    def scrape(self, dates=None):
        """
        Wrapper function to interface between swepy and nD

        Parameters
        ----------
        dates: List(datetime*)
            list of dates to scrape from
        """
        if self.local_session is False:
            if self.check_params() is False:
                return
        if dates is None:  # letting class choose all dates
            dates = self.dates
        for date in tqdm(dates):
            file19 = self.get_file(date, "19H")
            file37 = self.get_file(date, "37H")
            result1 = self.nD.download_file(**file19)
            if result1 is False:
                try:
                    result1 = self.nD.download_file(**file19)
                except:
                    print("failing on 19H {}".format(date))
                    pass
            result2 = self.nD.download_file(**file37)
            if result2 is False:
                try:
                    result2 = self.nD.download_file(**file37)
                except:
                    print("failing on 37H on {}".format(date))
                    pass
            self.down19list.append(result1[0])
            self.down37list.append(result2[0])
        return (result1, result2)

    @staticmethod
    def safe_subtract(tb19, tb37):
        """
        Check size of each file, often the 19 and 37
        matrices are one unit off of eachother.

        Chops the larger matrix to match the smaller matrix
        """

        shape1 = np.shape(tb19)
        shape2 = np.shape(tb37)
        s1 = [shape1[0], shape1[1], shape1[2]]
        s2 = [shape2[0], shape2[1], shape2[2]]
        if s1[1] < s2[1]:
            s2[1] = s1[1]
        elif s1[1] > s2[1]:
            s1[1] = s2[1]
        if s1[2] < s2[2]:
            s2[2] = s1[2]
        elif s1[2] > s2[2]:
            s1[2] = s2[2]
        tb19 = tb19[:, : s1[1] - 1, : s1[2] - 1]
        tb37 = tb37[:, : s2[1] - 1, : s2[2] - 1]
        tb = tb19 - tb37
        return tb

    def clean_dirs(self):
        """
        Delete files in directory
        Useful for cleaning up with repeated testing
        """

        os.chdir(self.wget)
        files = glob.glob("*")
        for f in files:
            os.remove(f)
        os.chdir(self.path19)
        files = glob.glob("*")
        for f in files:
            os.remove(f)
        os.chdir(self.path37)
        files = glob.glob("*")
        for f in files:
            os.remove(f)
        return

    def check_params(self):
        """
        Helper function to check that all the class members are set before
        attempting to web scrape or subset.

        Used by test suite and to check params are set before scraping.
        """

        proceed = True

        params = {
            "dates": self.dates,
            "bounding coordinates": self.geo_list,
            "grid": self.grid,
            "username": self.username,
            "password": self.password,
        }
        for key, value in params.items():
            if value is None:
                print("{} needs to be set by 'set_params'".format(key))
                proceed = False
        if proceed is False:
            print(
                "Please use the set_...() functions to set missing parameters,\
                    see the documentation for guidance"
            )
        return proceed


'''
    def plot_a_day(self, token, files=None, inday=None):
        """Plot swe subset with mapbox for confirmation of study area

        Takes final files from workflow, opens them, dumps info into geojson
        and plots via mapbox. It is recommended to use a jupyter notebook for viewing.

        Parameters
        ----------
        token: string
            mapbox token
        files: list
            list of file names ['19H', '37H']
        inday: int
            day of time series to plot on, defaults to 0
        """
        day = 0 if inday is None else inday
        # if no files passed, use final concatenated cubes
        if files is None:
            tb_19 = process.get_array(self.concatlist[0], "r", format="NETCDF4")
            tb_37 = process.get_array(self.concatlist[1], "r", format="NETCDF4")
        else:
            tb_19 = process.get_array(files[0], "r", format="NETCDF4")
            fid_37H = process.get_array(files[1], "r", format="NETCDF4")

        # extract x,y, and tb from cubes
        x = fid_19H.variables["x"][:]
        y = fid_19H.variables["y"][:]

        tb_19H = fid_19H.variables["TB"][:]
        tb_37H = fid_37H.variables["TB"][:]
        if self.high_res is True:  # downsample tb37
            tb_37H[tb_37H.mask] = 0.00001
            tb_37H = block_reduce(tb_37H, block_size=(1, 2, 2), func=np.mean)
            tb_37H = ma.masked_values(tb_37H, 0.00001)

        tb = self.safe_subtract(tb_19H, tb_37H)
        lats = np.zeros((len(y), len(x)), dtype=np.float64)
        lons = np.zeros((len(y), len(x)), dtype=np.float64)
        grid = easeReproject.EaseReproject(
            gridname=fid_19H.variables["crs"].long_name
        )
        one_day = tb[day, :, :]
        df = pd.DataFrame(columns=["lat", "lon", "swe"])
        for i, xi in enumerate(x):
            for j, yj in enumerate(y):
                row, col = grid.map_to_grid(xi, yj)
                lat, lon = grid.grid_to_geographic(row, col)
                lats[j, i] = lat
                lons[j, i] = lon
        for i in range(len(one_day[:, 1])):
            for j in range(len(one_day[1, :])):
                df = df.append(
                    {
                        "lat": lats[i][j],
                        "lon": lons[i][j],
                        "swe": one_day[i][j],
                    },
                    ignore_index=True,
                )
        os.chdir(self.working_dir)
        df_to_geojson(
            df,
            filename="swe_1day.geojson",
            properties=["swe"],
            lat="lat",
            lon="lon",
        )
        measure = "swe"
        color_breaks = [
            round(df[measure].quantile(q=x * 0.1), 2) for x in range(1, 9)
        ]
        color_stops = create_color_stops(color_breaks, colors="YlGnBu")
        # Create the viz from the dataframe
        viz = CircleViz(
            "swe_1day.geojson",
            access_token=token,
            color_property="swe",
            color_stops=color_stops,
            center=(self.center),
            zoom=3,
            below_layer="waterway-label",
        )

        viz.show()
'''



import os
from swepy.easeReproject import EaseReproject
import glob
import pytest


def test_ease_nogrid():
    with pytest.raises(ValueError):
        EaseReproject("From Sea to Summit")


def test_ease_incorrect_res():
    with pytest.raises(ValueError):
        EaseReproject("EASE2_N3.126km")


def test_ease_incorrect_proj():
    with pytest.raises(ValueError):
        EaseReproject("EASE3_W25km")


def test_grid_to_geo():
    row = 359.5
    col = 359.5
    n25g = EaseReproject("EASE2_N25km")
    assert n25g.grid_to_geographic(row, col) == pytest.approx((90.0, 0.0), 0.1)


def test_verbose():
    n25g = EaseReproject("EASE2_N25km", verbose=True)
    assert isinstance(n25g, EaseReproject)



import swepy.analysis as analysis
import pytest
import numpy as np
import pandas as pd
import matplotlib
import datetime


@pytest.fixture
def a():
    swe = np.zeros((3000, 50, 50))
    return analysis.Analysis(datetime.date(1993, 1, 1), swe)


def test_make_df_time():
    """
    Make sure timestamp is correctly created in dataframe
    """
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    t = a.make_df()
    print(type(t.time[0]))
    assert t.time[0] == pd.Timestamp("1993-01-01 00:00:00")


def test_make_df_type():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    t = a.make_df()
    assert type(t) == pd.DataFrame


def test_create_splits():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    years = a.create_year_splits()
    assert years == [0, 365, 730, 1095, 1460, 1826, 2191, 2556, 2921, 3000]


def test_count_melt_onset_mp():
    """
    Note: right now making zero matrix swe values
    - should maybe make fixture that has an instantiated analysis object with the swe cube inside
    """
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.count_melt_onset_mp()
    assert type(c) == pd.DataFrame


def test_mask_year_df():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    df = a.make_df()
    df = a.mask_year_df(df, 1994)
    assert df.iloc[0].time.year == 1994


def test_melt_date_year():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    df = a.make_df()
    counts = a.melt_date_year(df)
    assert type(counts) == dict


def test_count_index():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.count_melt_onset_index(swe)
    assert type(c) == pd.DataFrame


def test_summer_length():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.summer_length(swe)
    assert type(c) == dict


def test_summer_length2():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.summer_length(swe)
    assert type(c[(1, 1)]) == dict


def test_summer_diff1():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.summer_length(swe)
    diff, heatmap = a.summer_diff(c)
    if diff < 0.1:
        diff = 0
    assert diff == 0


def test_summer_diff2():
    swe = np.zeros((1000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.summer_length(swe)
    diff, heatmap = a.summer_diff(c)
    assert np.shape(heatmap) == (50, 50)


def test_display_diffmap():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    c = a.summer_length(swe)
    diff, heatmap = a.summer_diff(c)
    im = a.display_summer_change(True)
    assert type(im) == matplotlib.image.AxesImage


def test_display_melt():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    melt = a.melt_date_year(a.count_melt_onset_mp())
    fig = a.display_melt_onset_change(melt, 1993, 1995, True)
    assert type(fig) == matplotlib.figure.Figure


def test_display_melt_fail1():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    melt = a.melt_date_year(a.count_melt_onset_mp())
    with pytest.raises(Exception):
        a.display_melt_onset_change(melt, 1944, 1993)


def test_display_melt_fail2():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 1, 1), swe)
    melt = a.melt_date_year(a.count_melt_onset_mp())
    with pytest.raises(Exception):
        a.display_melt_onset_change(melt, 1993, 1944)


def test_create_splits_mid_year():
    swe = np.zeros((3000, 50, 50))
    a = analysis.Analysis(datetime.date(1993, 5, 1), swe)
    years = a.create_year_splits()
    assert years[0] == 121



import os
from swepy.nsidcDownloader import nsidcDownloader
import glob
import pytest
import datetime


def test_nD_local():
    file = {
        "protocol": "http",
        "server": "localhost:8000",
        "datapool": "MEASURES",
        "dataset": "NSIDC-0630",
        "version": "001",
        "projection": "EASE2",
        "resolution": "6.25km",
        "platform": "F17",
        "sensor": "SSMIS",
        "date1": datetime.date(2010, 1, 1),
        "date2": datetime.date(2010, 1, 1),
        "channel": "19H",
        "grid": "N",
        "pass": "M",
        "algorithm": "SIR",
        "input": "CSU",
        "dataversion": "v1.3",
    }
    nD = nsidcDownloader(no_auth=True)
    nD.download_file(**file)
    list1 = glob.glob("*.nc")
    assert (
        "NSIDC-0630-EASE2_N6.25km-F17_SSMIS-2010001-19H-M-SIR-CSU-v1.3.nc"
        in list1
    )


def test_get_auth():
    with pytest.raises(PermissionError):
        nsidcDownloader("room", "backend")


def test_file_not_found():
    with pytest.raises(FileNotFoundError):
        file = {
            "protocol": "http",
            "server": "localhost:8000",
            "datapool": "MEASURES",
            "dataset": "NSIDC-0630",
            "version": "001",
            "projection": "EASE2",
            "resolution": "6.25km",
            "platform": "F17",
            "sensor": "SSMIS",
            "date1": datetime.date(2010, 1, 1),
            "date2": datetime.date(2010, 1, 1),
            "channel": "19H",
            "grid": "K",
            "pass": "M",
            "algorithm": "SIR",
            "input": "CSU",
            "dataversion": "v1.3",
        }
        nD = nsidcDownloader(no_auth=True)
        nD.download_file(**file)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from multiprocessing import Pool, Process, cpu_count
import math


class Analysis:
    def __init__(self, start_date, swe):
        """
        Parameters
        ----------
        start_date: datetime
            Day to start building time series off of
            Infer length from number of days in data
        """
        self.leap_years = [1992, 1996, 2000, 2004, 2008, 2012, 2016]
        self.swe = swe
        self.start_date = start_date
        self.time = pd.date_range(
            start_date, periods=np.shape(self.swe)[0], freq="D"
        )
        # self.melt_df = self.make_df(time)
        self.year_splits = self.create_year_splits()

    def make_df(self, time=None, columns=["time", "count"]):
        """
        Given a time array, create dateframe with time and count columns

        Tool for melt onset workflow

        Parameters
        ----------
        time: pd.series
            time series of datetime.date() objects
        columns: list (Optional)
            column names to add in returned dataframe

        Returns
        -------
        df: pd.DataFrame
            dataframe with time and count for melt onset analysis
        """
        if time is None:
            time = self.time
        df = pd.DataFrame(columns=columns)
        df.time = time
        df["count"].values[:] = 0
        return df

    def create_year_splits(self):
        """
        Take time array from class and create an array of year split indexes

        Returns
        -------
        year_splits: list
            list of julian dates for the start of each year in time series
        """
        years = set(self.time.year)
        years = list(years)
        print("Years in time series: {}".format(years))
        tt = self.start_date.timetuple()
        if tt.tm_yday != 1:
            year_splits = [tt.tm_yday]
        else:
            year_splits = [0]
        for i, year in enumerate(years):
            if years[i - 1] in self.leap_years:
                year_splits.append(year_splits[-1] + 366)
            else:
                year_splits.append(year_splits[-1] + 365)
        if year_splits[-1] > np.shape(self.swe)[0]:
            year_splits[-1] = np.shape(self.swe)[0]
        return year_splits

    def count_melt_onset_index(self, swe):  # RENAME __COUNT
        """
        Count the date that each pixel reaches zero for first time of season on a given date.
        Useful for comparison between years of a given region.

        Parameters
        ----------
        swe: np.array
            swe cube (should be clean and relatively smooth)
        """
        melt_df = self.make_df(self.time)
        for i, year in enumerate(
            self.year_splits[0 : len(self.year_splits) - 1]
        ):
            # generate melt date boolean matrix
            bool_1 = np.zeros((np.shape(swe)[1], np.shape(swe)[2]), dtype=bool)
            # loop through the days of this year
            for d in range(self.year_splits[i], self.year_splits[i + 1] - 1):
                # loop through x and then y of the image
                for x in range(np.shape(swe)[1]):
                    for y in range(np.shape(swe)[2]):
                        if bool_1[x, y] is False:
                            if (
                                swe[d, x, y] == 0
                            ):  # less than 5mm is zero (error)
                                melt_df.loc[d, "count"] += 1
                                bool_1[x, y] = True
        return melt_df

    def count_melt_onset_mp(self):  # RENAME, get rid of mp
        """
        Count the date that each pixel reaches zero for first time of season on a given date.
        Useful for comparison between years of a given region.

        Helper function to manage multi-processing pool for counting
        melt onset date. Use __count to use a single core

        Parralelize on the 2nd axis (spatially)

        Returns
        -------
        df: pandas DataFrame
                Count of melt dates of every pixel in image
        """
        cpus = cpu_count()
        swe_parts = np.array_split(self.swe, cpus, axis=2)
        with Pool(cpus) as p:
            parts = p.map(self.count_melt_onset_index, swe_parts)
            df = parts[0]  # recombine parts
            for i in range(1, len(parts)):
                df["count"] = df["count"].add(parts[i]["count"])
            return df

    def mask_year_df(self, df, year):
        """
        Mask single year out of pandas dataframe based on "time" column

        Parameters
        ----------
        df: pd.DataFrame
            dataframe containing a "time" column
        year: datetime.date
            desired year to get from df
        """
        mask = df["time"].dt.year == year
        return df[mask]

    def melt_date_year(self, df):
        """
        Grab the counts of each year and stick them in own key of dictionary
        Makes for easy comparison of melt times each year.

        Parameters
        ----------
        df: pd.DataFrame
            dataframe with columns ["time", "count"]

        Returns
        -------
        counts_dict: dict
            dictionary with key=year : value=sum(count in year)
        """
        df_dict = {}
        counts_dict = {}
        for i in set(self.time.year):
            df_dict[i] = self.mask_year_df(df, i)
            counts_dict[i] = df_dict[i]["count"].values
        return counts_dict

    def summer_length(self, smooth_cube):
        """
        Function to track summer length for a pixel over each year
        Generates hash table of key:value = (x,y): summer length in days

        Parameters
        ----------
        smooth_cube: np.array
            smoothed (temporally) swe cube

        Returns
        -------
        store: dict
            dict using tuple (x,y) as hash key and list of summer lengths by year as key
        """
        shape = np.shape(smooth_cube)
        store = {}
        for x in range(shape[1]):
            for y in range(shape[2]):
                lengths = {}
                firstofyear = True
                f = 0
                lastofyear = True
                time = smooth_cube[:, x, y].tolist()
                for index, val in enumerate(time, start=1):
                    if index % 365 != 0:
                        if val == 0 and firstofyear:
                            f = index % 365
                            firstofyear = False
                        if val != 0 and firstofyear is False and lastofyear:
                            summerlength = ((index - 1) % 365) - f
                            lengths[math.floor((index / 365) + 1993)] = (
                                lengths.get(
                                    math.floor((index / 365) + 1993), 0
                                )
                                + summerlength
                            )
                            lastofyear = False
                    else:
                        firstofyear = True
                        lastofyear = True
                store[(x, y)] = lengths
        return store

    def summer_diff(self, summer_dict, smooth_cube=None):
        """
        Find the average change in summer length by pixel and overall

        Parameters
        ----------
        summer_dict: dict
            dictionary containing the summer lengths of each year by pixel
        smooth_cube: np.array(x,x,x)
            swe cube of clean swe data, OPTIONAL
        """
        if smooth_cube is None:
            smooth_cube = self.swe
        shape = np.shape(smooth_cube)
        diffmap = np.zeros((shape[1], shape[2]))
        total_diff = 0
        for key in summer_dict:  # grab new pixel dict
            keys = list(summer_dict[key].keys())
            diff = sum(
                [
                    summer_dict[keys[i]] - summer_dict[keys[i - 1]]
                    for i in range(1, len(keys))
                ]
            )
            diffmap[key[0], key[1]] = diff / 12
            total_diff += diff / 6
        self.diffmap = diffmap
        return total_diff / len(summer_dict.keys()), diffmap

    def display_summer_change(self, interactive=False):
        """
        Simple built in function for displaying the generated summer change heat map

        Parameters
        ----------
        interactive: bool
            boolean switch for unit testing to help close plots (would like to remove)
        """
        im = plt.imshow(self.diffmap)
        plt.colorbar()
        if interactive:
            plt.ion()
            plt.show()
            plt.pause(0.001)
            plt.close()
        else:
            plt.show()
        return im

    def display_melt_onset_change(self, dict, year1, year2, interactive=False):
        """
        CURRENTLY ONLY WORKS IF FULL YEARS ARE SCRAPED
        Given a dictionary of melt onset counts by date, viz two years against eachother

        Parameters
        ----------
        dict: dict
            dictionary of year:list(counts by year) pairs
        year1: datetime.date
            first year to plot on bar graph
        year2: datetime.date
            second year to plot on bar graph
        interactive: bool
            Unit testing parameter to help closing plots

        Returns
        -------
        fig: matplotlib.figure.Figure
            figure for future viz if desired
        """
        len1 = 366 if year1 in self.leap_years else 365
        len2 = 366 if year2 in self.leap_years else 365
        fig, ax = plt.subplots(1, 1, figsize=(15, 10))
        try:
            plt.bar(
                np.arange(len1),
                dict[year1],
                ec="black",
                width=1,
                zorder=2,
                label=str(year1),
            )
        except KeyError:
            print("Year 1 is not in the time series, please try again")
            raise KeyError
        try:
            plt.bar(
                np.arange(len2),
                dict[year2],
                ec="black",
                fc="red",
                width=1,
                zorder=4,
                alpha=0.50,
                label=str(year2),
            )
        except KeyError:
            print("Year 2 is not in the time series, please try again")
            raise KeyError
        plt.title(
            "Initial Zero SWE Date: {} and {}".format(str(year1), str(year2)),
            fontsize=20,
        )
        plt.xlabel("Day of Year", fontsize=16)
        plt.ylabel("Count of Pixels to Reach Zero", fontsize=16)
        plt.grid(0.25, zorder=1)
        plt.xlim([50, 200])
        plt.legend()
        if interactive:
            plt.ion()
            plt.show()
            plt.pause(0.001)
            plt.close()
        else:
            plt.show()
        return fig



import os
from http.server import HTTPServer as BaseHTTPServer, SimpleHTTPRequestHandler


class HTTPHandler(SimpleHTTPRequestHandler):
    """This handler uses server.base_path instead of always using os.getcwd()"""

    def translate_path(self, path):
        path = SimpleHTTPRequestHandler.translate_path(self, path)
        relpath = os.path.relpath(path, os.getcwd())
        fullpath = os.path.join(self.server.base_path, relpath)
        return fullpath


class HTTPServer(BaseHTTPServer):
    """The main server, you pass in base_path which is the path you want to serve requests from"""

    def __init__(
        self, base_path, server_address, RequestHandlerClass=HTTPHandler
    ):
        self.base_path = base_path
        BaseHTTPServer.__init__(self, server_address, RequestHandlerClass)


web_dir = os.path.join(os.path.dirname(__file__), "data")
httpd = HTTPServer(web_dir, ("", 8000))
httpd.serve_forever()

# serves files on localhost:8000/file_name



from swepy.pipeline import Swepy
import swepy.process as process
import datetime
import os
import pytest
import numpy as np


@pytest.fixture
def scraped_files():
    date = datetime.date(2010, 1, 1)
    s1 = Swepy(os.getcwd(), ul="N", lr="N")
    s1.set_dates(date, date)
    s1.set_login("test", "test")
    files = s1.scrape()
    return (files[0][0], files[1][0])


@pytest.fixture
def arrays(scraped_files):
    tb19 = process.get_array(scraped_files[0])
    tb37 = process.get_array(scraped_files[1])
    return (tb19, tb37)


@pytest.fixture
def swe(arrays):
    tb19 = process.vector_clean(arrays[0])
    tb37 = process.vector_clean(arrays[1])
    swe = Swepy.safe_subtract(tb19=tb19, tb37=tb37)
    return swe


def test_get_array(arrays):
    assert type(arrays[0]) == np.ma.core.MaskedArray


def test_vector_clean(arrays):
    cleantb19 = process.vector_clean(arrays[0])
    bool1 = True
    if not np.isnan(cleantb19.all()):
        bool1 = False
    assert bool1 is False


def test_pandas_fill(arrays):
    clean = process.pandas_fill(arrays[0][:, 1, 1])
    bool1 = True
    if not np.isnan(clean).all():
        bool1 = False
    assert bool1 is False


def test_apply_filter_success(arrays):
    tb19 = process.vector_clean(arrays[0])
    tb37 = process.vector_clean(arrays[1])
    swe = Swepy.safe_subtract(tb19=tb19, tb37=tb37)
    swe = np.concatenate(
        (
            np.concatenate((np.concatenate((swe, swe), axis=0), swe), axis=0),
            swe,
        ),
        axis=0,
    )
    swe = swe[:, 1:2, 1:2]
    swe = process.apply_filter(swe)
    assert swe.min() == 0


def test_apply_filter_fail(arrays):
    tb19 = process.vector_clean(arrays[0])
    clean19 = process.apply_filter(tb19)
    assert clean19 == ValueError


def test_apply_filter_large():
    tb19 = np.zeros((100, 5, 5))
    clean19 = process.apply_filter(tb19)
    assert np.shape(clean19) == (100, 5, 5)


def test_apply_filter_mp():
    tb19 = np.zeros((100, 50, 50))
    clean19 = process.apply_filter_mphelper(tb19)
    assert np.shape(clean19) == (100, 50, 50)


def test_apply_filter_mp_fail():
    tb19 = np.zeros((100, 1))
    with pytest.raises(Exception):
        process.apply_filter_mphelper(tb19)


def test_govf():
    np.random.seed(1)
    array = np.random.rand(1, 10, 10)
    res = process.govf(array.ravel(), 4)
    assert res == 0.9476204544544236


def test_ocean_mask():
    array = np.ones((1, 142, 130))
    arr = process.mask_ocean_winter(array)
    assert type(arr) is np.ndarray
